"""Where command handling and logging reside. This also is where I wrote extensions in commands.py"""
from .command import (Command, CommandCollection, Context, Event, Extender,
                      Extension, ExtensionCollection)
from .logging import logging